
CREATE TABLE audit (
    audit_id          SERIAL PRIMARY KEY,
    timestamp         TIMESTAMPTZ NOT NULL DEFAULT CURRENT_TIMESTAMP,
    app_user_id       INTEGER,    -- would link to app_user table
    audit_type_id     INTEGER     -- would link to audit_type table
);

CREATE TABLE person (
    person_id         SERIAL PRIMARY KEY,
    name              TEXT NOT NULL,
    date_of_birth     DATE NOT NULL,
    address           TEXT,
    audit_id          INTEGER REFERENCES audit(audit_id)
);

CREATE TABLE person_hst (
    person_hst_id     SERIAL PRIMARY KEY,
    person_id         INTEGER,
    name              TEXT NOT NULL,
    date_of_birth     DATE NOT NULL,
    address           TEXT,
    audit_id          INTEGER REFERENCES audit(audit_id),
    effective_from    TIMESTAMPTZ,
    effective_until   TIMESTAMPTZ,
    removed           BOOLEAN NOT NULL DEFAULT FALSE
);


CREATE OR REPLACE FUNCTION fi_person()
    RETURNS trigger
    LANGUAGE plpgsql
AS $function$
BEGIN
    -- Insert current data into history
    INSERT INTO person_hst (
        effective_until,
        effective_from,
        address,
        audit_id,
        date_of_birth,
        person_id,
        name
    ) SELECT
        'infinity' AS effective_until,
        audit.timestamp,
        NEW.address,
        NEW.audit_id,
        NEW.date_of_birth,
        NEW.person_id,
        NEW.name
    FROM  audit
    WHERE audit_id = NEW.audit_id;

    RETURN NEW;
END;
$function$
;


CREATE OR REPLACE FUNCTION fu_person()
    RETURNS trigger
    LANGUAGE plpgsql
AS $function$
BEGIN
    IF NEW.audit_id = OLD.audit_id THEN
        -- Update current data in history
        UPDATE person_hst
        SET
            address = NEW.address,
            date_of_birth = NEW.date_of_birth,
            person_id = NEW.person_id,
            name = NEW.name
        WHERE person_hst.person_id = NEW.person_id
        AND   person_hst.audit_id = NEW.audit_id
        AND   person_hst.effective_until = 'infinity';
    ELSE
        -- Update the end_date on the previous record
        UPDATE person_hst
        SET effective_until = (
            SELECT timestamp
            FROM   audit
            WHERE  audit.audit_id = NEW.audit_id
        )
        WHERE person_hst.person_id = NEW.person_id
        AND   effective_until  = 'infinity';

        -- Insert current data into history
        INSERT INTO person_hst (
            effective_until,
            effective_from,
            address,
            audit_id,
            date_of_birth,
            person_id,
            name
        ) SELECT
            'infinity' AS effective_until,
            audit.timestamp,
            NEW.address,
            NEW.audit_id,
            NEW.date_of_birth,
            NEW.person_id,
            NEW.name
        FROM  audit
        WHERE audit_id = NEW.audit_id;
    END IF;

    RETURN NEW;
END;
$function$
;


CREATE OR REPLACE FUNCTION fd_person()
    RETURNS trigger
    LANGUAGE plpgsql
AS $function$
BEGIN
    -- Confirm that history shows FR_person has been called
    IF NOT exists (
        SELECT * FROM person_hst h
        WHERE h.person_id = OLD.person_id
        AND h.effective_until = 'infinity'
        AND removed
    ) THEN
        RAISE EXCEPTION 'Use fr_person() to delete person records';
    END IF;
    -- Allow the delete to proceed
    RETURN OLD;
END;
$function$
;


CREATE OR REPLACE FUNCTION fr_person(person, bigint)
    RETURNS boolean
    LANGUAGE plpgsql
AS $function$
BEGIN
    -- Update the end_date on the previous record
    UPDATE person_hst
    SET    effective_until  = (
        SELECT timestamp
        FROM   audit
        WHERE  audit.audit_id = $2
    )
    WHERE  person_hst.person_id = $1.person_id
    AND    effective_until  = 'infinity';

    -- Insert the new HST record
    INSERT INTO person_hst (
        effective_until,
        effective_from,
        removed,
        audit_id,
        address,
        date_of_birth,
        person_id,
        name
    ) SELECT
        'infinity' AS effective_until,
        audit.timestamp,
        TRUE,
        $2,
        $1.address,
        $1.date_of_birth,
        $1.person_id,
        $1.name
    FROM   audit
    WHERE  audit.audit_id = $2;

    -- Delete the current record
    DELETE FROM person
    WHERE  person.person_id = $1.person_id;

    RETURN TRUE;
END;
$function$
;


CREATE TRIGGER ti_person
    AFTER INSERT
    ON person
    FOR EACH ROW EXECUTE PROCEDURE fi_person();


CREATE TRIGGER tu_person
    AFTER UPDATE
    ON person
    FOR EACH ROW EXECUTE PROCEDURE fu_person();


CREATE TRIGGER td_person
    BEFORE DELETE
    ON person
    FOR EACH ROW EXECUTE PROCEDURE fd_person();

